import ctypes
from kanzi_c_api import _lib, _libc, cContext_p, dContext_p, cData, dData


class KanziError(RuntimeError):
    pass


def _check(rc, msg):
    if rc != 0:
        raise KanziError(f"{msg} (error code {rc})")


# -----------------------------------------------------------------------------
# Compressor
# -----------------------------------------------------------------------------

class Compressor:
    def __init__(
        self,
        dst_path,
        transform=b"LZ",
        entropy=b"Huffman",
        block_size=1 << 20,
        jobs=1,
        checksum=0,
        headerless=0,
    ):
        self._file = _libc.fopen(dst_path.encode(), b"wb")
        if not self._file:
            raise OSError("fopen failed")

        self._ctx = cContext_p()

        params = cData()
        params.transform = transform
        params.entropy = entropy
        params.blockSize = block_size
        params.jobs = jobs
        params.checksum = checksum
        params.headerless = headerless

        rc = _lib.initCompressor(
            ctypes.byref(params),
            self._file,
            ctypes.byref(self._ctx),
        )
        _check(rc, "initCompressor failed")

    def compress(self, data: bytes) -> int:
        src = (ctypes.c_ubyte * len(data)).from_buffer_copy(data)
        out_size = ctypes.c_size_t(0)

        rc = _lib.compress(
            self._ctx,
            src,
            len(data),
            ctypes.byref(out_size),
        )
        _check(rc, "compress failed")
        return out_size.value

    def close(self) -> int:
        out_size = ctypes.c_size_t(0)
        rc = _lib.disposeCompressor(
            ctypes.byref(self._ctx),
            ctypes.byref(out_size),
        )
        _check(rc, "disposeCompressor failed")

        _libc.fclose(self._file)
        self._file = None
        return out_size.value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()


# -----------------------------------------------------------------------------
# Decompressor
# -----------------------------------------------------------------------------

class Decompressor:
    def __init__(
        self,
        src_path,
        buffer_size,
        jobs=1,
        headerless=0,
        **headerless_params,
    ):
        self._file = _libc.fopen(src_path.encode(), b"rb")
        if not self._file:
            raise OSError("fopen failed")

        self._ctx = dContext_p()

        params = dData()
        params.bufferSize = buffer_size
        params.jobs = jobs
        params.headerless = headerless

        if headerless:
            params.transform = headerless_params["transform"]
            params.entropy = headerless_params["entropy"]
            params.blockSize = headerless_params["blockSize"]
            params.originalSize = headerless_params["originalSize"]
            params.checksum = headerless_params["checksum"]
            params.bsVersion = headerless_params["bsVersion"]

        rc = _lib.initDecompressor(
            ctypes.byref(params),
            self._file,
            ctypes.byref(self._ctx),
        )
        _check(rc, "initDecompressor failed")

    def decompress_block(self, max_output: int) -> bytes:
        dst = (ctypes.c_ubyte * max_output)()
        in_size = ctypes.c_size_t(0)
        out_size = ctypes.c_size_t(max_output)

        rc = _lib.decompress(
            self._ctx,
            dst,
            ctypes.byref(in_size),
            ctypes.byref(out_size),
        )
        _check(rc, "decompress failed")

        return bytes(dst[: out_size.value])

    def close(self):
        rc = _lib.disposeDecompressor(ctypes.byref(self._ctx))
        _check(rc, "disposeDecompressor failed")
        _libc.fclose(self._file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()




import ctypes
import sys
import os

# -----------------------------------------------------------------------------
# Platform detection
# -----------------------------------------------------------------------------

if sys.platform.startswith("win"):
    KANZI_LIB_NAME = "kanzi.dll"
    LIBC_NAME = "msvcrt.dll"
elif sys.platform == "darwin":
    KANZI_LIB_NAME = "libkanzi.dylib"
    LIBC_NAME = "libc.dylib"
else:
    KANZI_LIB_NAME = "libkanzi.so"
    LIBC_NAME = "libc.so.6"

# -----------------------------------------------------------------------------
# Load shared libraries
# -----------------------------------------------------------------------------

_lib = ctypes.CDLL(KANZI_LIB_NAME)
_libc = ctypes.CDLL(LIBC_NAME)

# -----------------------------------------------------------------------------
# libc FILE*
# -----------------------------------------------------------------------------

FILE_p = ctypes.c_void_p

_libc.fopen.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
_libc.fopen.restype = FILE_p

_libc.fclose.argtypes = [FILE_p]
_libc.fclose.restype = ctypes.c_int

# -----------------------------------------------------------------------------
# Opaque contexts
# -----------------------------------------------------------------------------

class cContext(ctypes.Structure):
    pass


class dContext(ctypes.Structure):
    pass


cContext_p = ctypes.POINTER(cContext)
dContext_p = ctypes.POINTER(dContext)

# -----------------------------------------------------------------------------
# Compression parameters
# -----------------------------------------------------------------------------

class cData(ctypes.Structure):
    _fields_ = [
        ("transform", ctypes.c_char * 64),
        ("entropy", ctypes.c_char * 16),
        ("blockSize", ctypes.c_size_t),
        ("jobs", ctypes.c_uint),
        ("checksum", ctypes.c_int),
        ("headerless", ctypes.c_int),
    ]


class dData(ctypes.Structure):
    _fields_ = [
        # Required
        ("bufferSize", ctypes.c_size_t),
        ("jobs", ctypes.c_uint),
        ("headerless", ctypes.c_int),

        # Headerless-only
        ("transform", ctypes.c_char * 64),
        ("entropy", ctypes.c_char * 16),
        ("blockSize", ctypes.c_uint),
        ("originalSize", ctypes.c_size_t),
        ("checksum", ctypes.c_int),
        ("bsVersion", ctypes.c_int),
    ]

# -----------------------------------------------------------------------------
# Function prototypes - Compressor
# -----------------------------------------------------------------------------

_lib.getCompressorVersion.argtypes = []
_lib.getCompressorVersion.restype = ctypes.c_uint

_lib.initCompressor.argtypes = [
    ctypes.POINTER(cData),
    FILE_p,
    ctypes.POINTER(cContext_p),
]
_lib.initCompressor.restype = ctypes.c_int

_lib.compress.argtypes = [
    cContext_p,
    ctypes.POINTER(ctypes.c_ubyte),
    ctypes.c_size_t,
    ctypes.POINTER(ctypes.c_size_t),
]
_lib.compress.restype = ctypes.c_int

_lib.disposeCompressor.argtypes = [
    ctypes.POINTER(cContext_p),
    ctypes.POINTER(ctypes.c_size_t),
]
_lib.disposeCompressor.restype = ctypes.c_int

# -----------------------------------------------------------------------------
# Function prototypes - Decompressor
# -----------------------------------------------------------------------------

_lib.getDecompressorVersion.argtypes = []
_lib.getDecompressorVersion.restype = ctypes.c_uint

_lib.initDecompressor.argtypes = [
    ctypes.POINTER(dData),
    FILE_p,
    ctypes.POINTER(dContext_p),
]
_lib.initDecompressor.restype = ctypes.c_int

_lib.decompress.argtypes = [
    dContext_p,
    ctypes.POINTER(ctypes.c_ubyte),
    ctypes.POINTER(ctypes.c_size_t),
    ctypes.POINTER(ctypes.c_size_t),
]
_lib.decompress.restype = ctypes.c_int

_lib.disposeDecompressor.argtypes = [
    ctypes.POINTER(dContext_p),
]
_lib.disposeDecompressor.restype = ctypes.c_int

# -----------------------------------------------------------------------------
# Optional helpers (recommended for kanzi.py)
# -----------------------------------------------------------------------------

def fopen(path: bytes, mode: bytes) -> FILE_p:
    return _libc.fopen(path, mode)


def fclose(fp: FILE_p) -> int:
    return _libc.fclose(fp)


# -----------------------------------------------------------------------------
# Public exports
# -----------------------------------------------------------------------------

__all__ = [
    # libraries
    "_lib",
    "_libc",

    # FILE*
    "FILE_p",
    "fopen",
    "fclose",

    # contexts
    "cContext",
    "dContext",
    "cContext_p",
    "dContext_p",

    # params
    "cData",
    "dData",

    # functions (via _lib)
]




import os
import tempfile

from kanzi import Compressor, Decompressor, KanziError


# -----------------------------------------------------------------------------
# Utilities
# -----------------------------------------------------------------------------

def make_params():
    return dict(
        transform=b"LZX",
        entropy=b"HUFFMAN",
        block_size=1024,
        jobs=1,
        checksum=0,
        headerless=0,
    )


def fill_buffer(size):
    return bytes((i * 17 + 3) & 0xFF for i in range(size))


def write_file(path, data: bytes):
    with open(path, "wb") as f:
        f.write(data)


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------

def test_init_invalid():
    print("TEST: initCompressor invalid params...")

    # Our Python wrapper validates parameters eagerly,
    # so most invalid cases are raised as Python exceptions.

    try:
        Compressor(
            dst_path="/dev/null",
            transform=None,   # invalid
        )
        assert False, "init should fail on invalid transform"
    except Exception:
        pass


def test_init_dispose():
    print("TEST: init + dispose...")

    with tempfile.NamedTemporaryFile(delete=True) as tmp:
        with Compressor(tmp.name, **make_params()) as c:
            assert c is not None


def test_compress_small():
    print("TEST: small compression...")

    data = fill_buffer(256)

    with tempfile.NamedTemporaryFile(delete=True) as tmp:
        with Compressor(tmp.name, **make_params()) as c:
            written = c.compress(data)
            assert written >= 0


def test_compress_too_big():
    print("TEST: oversized block handling...")

    params = make_params()
    params["block_size"] = 1024

    big = fill_buffer(4096)

    with tempfile.NamedTemporaryFile(delete=True) as tmp:
        with Compressor(tmp.name, **params) as c:
            try:
                c.compress(big)
                assert False, "compress should fail on oversized input"
            except KanziError:
                pass


def test_compress_two_blocks():
    print("TEST: two-block compression...")

    params = make_params()
    params["block_size"] = 1024

    a = fill_buffer(300)
    b = fill_buffer(500)

    with tempfile.NamedTemporaryFile(delete=True) as tmp:
        with Compressor(tmp.name, **params) as c:
            out1 = c.compress(a)
            out2 = c.compress(b)
            assert out1 >= 0
            assert out2 >= 0


def test_basic_decompression():
    print("TEST: basic decompression...")

    input_data = b"Hello Kanzi! Hello Compression!"

    with tempfile.NamedTemporaryFile(delete=False) as comp:
        comp_name = comp.name

    # Compress
    with Compressor(
        comp_name,
        transform=b"LZ",
        entropy=b"ANS0",
        block_size=1 << 16,
        jobs=1,
        checksum=32,
        headerless=0,
    ) as c:
        c.compress(input_data)

    # Decompress
    with Decompressor(
        comp_name,
        buffer_size=1 << 16,
        jobs=1,
        headerless=0,
    ) as d:
        out = d.decompress_block(1024)

    assert out == input_data, "decompressed data mismatch"

    os.remove(comp_name)


def test_large_multi_block():
    print("TEST: large multi blocks")

    size = 2 * 1024 * 1024
    data = bytes((i * 7) & 0xFF for i in range(size))

    with tempfile.NamedTemporaryFile(delete=False) as comp:
        comp_name = comp.name

    # Compress
    with Compressor(
        comp_name,
        transform=b"LZ",
        entropy=b"FPAQ",
        block_size=256 * 1024,
        jobs=1,
        checksum=64,
        headerless=0,
    ) as c:
        offset = 0
        while offset < size:
            chunk = min(256 * 1024, size - offset)
            c.compress(data[offset:offset + chunk])
            offset += chunk

    # Decompress
    out = bytearray()

    with Decompressor(
        comp_name,
        buffer_size=256 * 1024,
        jobs=1,
        headerless=0,
    ) as d:
        while True:
            try:
                block = d.decompress_block(256 * 1024)
                if not block:
                    break
                out.extend(block)
            except KanziError:
                break  # expected EOF

    assert bytes(out) == data, "large decompression mismatch"

    os.remove(comp_name)


def test_headerless():
    print("TEST: headerless")

    input_data = b"HEADERLESS MODE IS ACTIVE"

    with tempfile.NamedTemporaryFile(delete=False) as comp:
        comp_name = comp.name

    # Compress (headerless)
    with Compressor(
        comp_name,
        transform=b"LZ",
        entropy=b"ANS0",
        block_size=1 << 15,
        jobs=1,
        checksum=0,
        headerless=1,
    ) as c:
        c.compress(input_data)

    # Decompress (headerless)
    with Decompressor(
        comp_name,
        buffer_size=1 << 15,
        jobs=1,
        headerless=1,
        transform=b"LZ",
        entropy=b"ANS0",
        blockSize=1 << 15,
        originalSize=len(input_data),
        checksum=0,
        bsVersion=1,
    ) as d:
        out = d.decompress_block(256)

    assert out == input_data, "headerless decompression mismatch"

    os.remove(comp_name)


# -----------------------------------------------------------------------------
# Main
# -----------------------------------------------------------------------------

def main():
    test_init_invalid()
    test_init_dispose()
    test_compress_small()
    test_compress_too_big()
    test_compress_two_blocks()

    test_basic_decompression()
    test_large_multi_block()
    test_headerless()

    print("All Python API tests passed.")


if __name__ == "__main__":
    main()

